
/*
    pbrt source code is Copyright(c) 1998-2016
                        Matt Pharr, Greg Humphreys, and Wenzel Jakob.

    This file is part of pbrt.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

    - Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    - Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
    IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
    TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */


// core/primitive.cpp*
#include "primitive.h"
#include "light.h"
#include "interaction.h"
#include "stats.h"

namespace pbrt {

STAT_MEMORY_COUNTER("Memory/Primitives", primitiveMemory);

// Primitive Method Definitions
Primitive::~Primitive() {}
const AreaLight *Aggregate::GetAreaLight() const
{
    #if LOGGING
        LOG(FATAL) <<
            "Aggregate::GetAreaLight() method"
            "called; should have gone to GeometricPrimitive";
    #endif
    return nullptr;
}

const Material *Aggregate::GetMaterial() const
{
    #if LOGGING
        LOG(FATAL) <<
            "Aggregate::GetMaterial() method"
            "called; should have gone to GeometricPrimitive";
    #endif
    return nullptr;
}

void Aggregate::ComputeScatteringFunctions(SurfaceInteraction *isect,
                                           MemoryArena &arena,
                                           TransportMode mode,
                                           bool allowMultipleLobes) const
{
    #if LOGGING
        LOG(FATAL) <<
            "Aggregate::ComputeScatteringFunctions() method"
            "called; should have gone to GeometricPrimitive";
    #endif
}

void Primitive::ComputeScatteringFunctions_Cor
(
    SurfaceInteraction_CorBSDF *isect,
    MemoryArena &arena,
    TransportMode mode,
    bool allowMultipleLobes
) const
{
    return ComputeScatteringFunctions(isect, arena, mode, allowMultipleLobes);
}

void Primitive::ComputeScatteringFunctions_CorDeb
(
    SurfaceInteraction_CorDebBSDF *isect,
    Float h,
    Float h1,
    Float h2,
    MemoryArena &arena,
    TransportMode mode,
    bool allowMultipleLobes
) const
{
    return ComputeScatteringFunctions(isect, arena, mode, allowMultipleLobes);
}

// TransformedPrimitive Method Definitions
TransformedPrimitive::TransformedPrimitive(std::shared_ptr<Primitive> &primitive,
                                           const AnimatedTransform &PrimitiveToWorld)
    : primitive(primitive), PrimitiveToWorld(PrimitiveToWorld)
{
    primitiveMemory += sizeof(*this);
}

bool TransformedPrimitive::Intersect(const Ray &r,
                                     SurfaceInteraction *isect) const {
    // Compute _ray_ after transformation by _PrimitiveToWorld_
    Transform InterpolatedPrimToWorld;
    PrimitiveToWorld.Interpolate(r.time, &InterpolatedPrimToWorld);
    Ray ray = Inverse(InterpolatedPrimToWorld)(r);
    if (!primitive->Intersect(ray, isect)) return false;
    r.tMax = ray.tMax;
    // Transform instance's intersection data to world space
    if (!InterpolatedPrimToWorld.IsIdentity())
        *isect = InterpolatedPrimToWorld(*isect);
    #if LOGGING
        CHECK_GE(Dot(isect->n, isect->shading.n), 0);
    #endif
    return true;
}

bool TransformedPrimitive::IntersectP(const Ray &r) const {
    Transform InterpolatedPrimToWorld;
    PrimitiveToWorld.Interpolate(r.time, &InterpolatedPrimToWorld);
    Transform InterpolatedWorldToPrim = Inverse(InterpolatedPrimToWorld);
    return primitive->IntersectP(InterpolatedWorldToPrim(r));
}

// GeometricPrimitive Method Definitions
GeometricPrimitive::GeometricPrimitive
(
    const std::shared_ptr<Shape> &shape,
    const std::shared_ptr<Material> &material,
    const std::shared_ptr<AreaLight> &areaLight,
    const MediumInterface &mediumInterface
)
    : shape(shape),
      material(material),
      areaLight(areaLight),
      mediumInterface(mediumInterface)
{
    primitiveMemory += sizeof(*this);
}

Bounds3f GeometricPrimitive::WorldBound() const { return shape->WorldBound(); }

bool GeometricPrimitive::IntersectP(const Ray &r) const {
    return shape->IntersectP(r);
}

bool GeometricPrimitive::Intersect(const Ray &r,
                                   SurfaceInteraction *isect) const
{
    Float tHit;
    if (!shape->Intersect(r, &tHit, isect)) return false;
    r.tMax = tHit;
    isect->primitive = this;
    #if LOGGING
        CHECK_GE(Dot(isect->n, isect->shading.n), 0.);
    #endif
    // Initialize _SurfaceInteraction::mediumInterface_ after _Shape_
    // intersection
    if (mediumInterface.IsMediumTransition())
        isect->mediumInterface = mediumInterface;
    else
        isect->mediumInterface = MediumInterface(r.medium);
    return true;
}

const AreaLight *GeometricPrimitive::GetAreaLight() const {
    return areaLight.get();
}

const Material *GeometricPrimitive::GetMaterial() const {
    return material.get();
}

void GeometricPrimitive::ComputeScatteringFunctions(SurfaceInteraction *isect,
                                                    MemoryArena &arena,
                                                    TransportMode mode,
                                                    bool allowMultipleLobes) const
{
    ProfilePhase p(Prof::ComputeScatteringFuncs);

    if (material) material->ComputeScatteringFunctions(isect,
                                                       arena,
                                                       mode,
                                                       allowMultipleLobes);

    #if LOGGING
        CHECK_GE(Dot(isect->n, isect->shading.n), 0.);
    #endif
}

void GeometricPrimitive::ComputeScatteringFunctions_Cor
(
    SurfaceInteraction_CorBSDF *isect,
    MemoryArena &arena,
    TransportMode mode,
    bool allowMultipleLobes
) const
{
    ProfilePhase p(Prof::ComputeScatteringFuncs);

    // std::cout << "computing scattering cor - geom prim" << std::endl;

    if (material)
    {
        // std::cout << "material defined" << std::endl;

        material->ComputeScatteringFunctions_Cor(isect,
                                                 arena,
                                                 mode,
                                                 allowMultipleLobes);
    }
}

void GeometricPrimitive::ComputeScatteringFunctions_CorDeb
(
    SurfaceInteraction_CorDebBSDF *isect,
    Float h,
    Float h1,
    Float h2,
    MemoryArena &arena,
    TransportMode mode,
    bool allowMultipleLobes
) const
{
    ProfilePhase p(Prof::ComputeScatteringFuncs);

    if (material)
    {
        material->ComputeScatteringFunctions_CorDeb(isect,
                                                    h,
                                                    h1,
                                                    h2,
                                                    arena,
                                                    mode,
                                                    allowMultipleLobes);
    }
}

void GeometricPrimitive::ComputeScatteringFunctions(SurfaceInteraction *isect,
                                                    MemoryArena &arena,
                                                    TransportMode mode,
                                                    bool allowMultipleLobes,
                                                    int depth) const
{
    ProfilePhase p(Prof::ComputeScatteringFuncs);

    if (material) material->ComputeScatteringFunctions(isect,
                                                       arena,
                                                       mode,
                                                       allowMultipleLobes,
                                                       depth);

    #if LOGGING
        CHECK_GE(Dot(isect->n, isect->shading.n), 0.);
    #endif
}

}  // namespace pbrt
