
/*
    pbrt source code is Copyright(c) 1998-2016
                        Matt Pharr, Greg Humphreys, and Wenzel Jakob.

    This file is part of pbrt.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

    - Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    - Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
    IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
    TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */


// samplers/random.cpp*
#include "samplers/random.h"
#include "paramset.h"
#include "sampling.h"
#include "stats.h"
#include "camera.h"

namespace pbrt {

RandomSampler::RandomSampler(int ns, int seed) : Sampler(ns), rng(seed) {}

Float RandomSampler::Get1D()
{
    ProfilePhase _(Prof::GetSample);
    #if LOGGING
        CHECK_LT(currentPixelSampleIndex, samplesPerPixel);
    #endif
    return rng.UniformFloat();
}

Point2f RandomSampler::Get2D()
{
    ProfilePhase _(Prof::GetSample);
    #if LOGGING
        CHECK_LT(currentPixelSampleIndex, samplesPerPixel);
    #endif
    return {rng.UniformFloat(), rng.UniformFloat()};
}

std::unique_ptr<Sampler> RandomSampler::Clone(int seed) {
    RandomSampler *rs = new RandomSampler(*this);
    rs->rng.SetSequence(seed);
    return std::unique_ptr<Sampler>(rs);
}

void RandomSampler::StartPixel(const Point2i &p)
{
    ProfilePhase _(Prof::StartPixel);
    for (size_t i = 0; i < sampleArray1D.size(); ++i)
        for (size_t j = 0; j < sampleArray1D[i].size(); ++j)
            sampleArray1D[i][j] = rng.UniformFloat();

    for (size_t i = 0; i < sampleArray2D.size(); ++i)
        for (size_t j = 0; j < sampleArray2D[i].size(); ++j)
            sampleArray2D[i][j] = {rng.UniformFloat(), rng.UniformFloat()};
    Sampler::StartPixel(p);
}

#if CAMERA_CENTERED

CameraSample RandomSampler::GetCameraSample(const Point2i &pRaster)
{
    CameraSample cs;
    cs.pFilm = (Point2f)(pRaster) + Point2f(0.5, 0.5);
    cs.time = Get1D();
    cs.pLens = Get2D();
    return cs;
}

#else

CameraSample RandomSampler::GetCameraSample(const Point2i &pRaster)
{
    CameraSample cs;
    cs.pFilm = (Point2f)pRaster + Get2D();
    cs.time = Get1D();
    cs.pLens = Get2D();
    return cs;
}

#endif

Sampler *CreateRandomSampler(const ParamSet &params)
{
    int ns = params.FindOneInt("pixelsamples", 4);
    return new RandomSampler(ns);
}

Sampler* CreateRandomSampler(int seed, int ns)
{
    std::cout << "BAD" << std::endl;
    return new RandomSampler(ns, seed);
}

}  // namespace pbrt
